import {
	GetServerSideProps,
	NextApiHandler,
	NextApiRequest as DefaultNextApiRequest,
	PageConfig,
} from 'next';
import { deserialize } from 'seroval';
import {
	NextApiResponse,
	NextApiRequest,
} from '../common/types';
import {
	ENCTYPE_APPLICATION_JSON,
	ENCTYPE_APPLICATION_OCTET_STREAM,
} from '../common/enctypes';
import { getBody } from '../utils/request';
import {
	CookieManager,
	BODY_COOKIE_KEY,
	STATUS_CODE_COOKIE_KEY,
	STATUS_MESSAGE_COOKIE_KEY,
	CONTENT_TYPE_COOKIE_KEY,
} from '../utils/cookies';
import { deserializeBody, EncTypeDeserializerMap } from '../utils/serialization';
import { IceformNextServerResponse } from './response';
import {
	DEFAULT_METHOD,
	DEFAULT_ENCODING,
	DEFAULT_RESPONSE_STATUS_CODE,
	ACTION_STATUS_CODE,
	METHODS_WITH_BODY,
	PREVENT_REDIRECT_FORM_KEY,
} from './constants';

export namespace destination {
	export const getServerSideProps = (
		gspFn?: GetServerSideProps,
	): GetServerSideProps => async (ctx) => {
		const req: NextApiRequest = {
			query: {
				...ctx.query,
				...(ctx.params ?? {}),
			},
			body: null,
		};
		const { method = DEFAULT_METHOD } = ctx.req;

		if (METHODS_WITH_BODY.includes(method.toUpperCase() as typeof METHODS_WITH_BODY[number])) {
			const body = await getBody(ctx.req);
			req.body = body.toString(DEFAULT_ENCODING);
		}

		const cookieManager = new CookieManager(ctx);
		// TODO how to properly remove cookies without leftovers?
		if (cookieManager.hasCookie(STATUS_CODE_COOKIE_KEY)) {
			ctx.res.statusCode = Number(
				cookieManager.getCookie(STATUS_CODE_COOKIE_KEY) || DEFAULT_RESPONSE_STATUS_CODE,
			);
			cookieManager.unsetCookie(STATUS_CODE_COOKIE_KEY);
		}

		if (cookieManager.hasCookie(STATUS_MESSAGE_COOKIE_KEY)) {
			ctx.res.statusMessage = cookieManager.getCookie(STATUS_MESSAGE_COOKIE_KEY) || '';
			cookieManager.unsetCookie(STATUS_MESSAGE_COOKIE_KEY);
		}

		const res: NextApiResponse = {};
		if (cookieManager.hasCookie(BODY_COOKIE_KEY)) {
			const resBody = cookieManager.getCookie(BODY_COOKIE_KEY);
			const contentType = cookieManager.getCookie(CONTENT_TYPE_COOKIE_KEY);
			if (contentType === ENCTYPE_APPLICATION_JSON) {
				res.body = deserialize(resBody);
			} else if (contentType === ENCTYPE_APPLICATION_OCTET_STREAM) {
				res.body = deserialize(resBody);
			} else {
				const c = console;
				c.warn('Could not parse response body, returning nothing');
				res.body = null;
			}
			cookieManager.unsetCookie(BODY_COOKIE_KEY);
			cookieManager.unsetCookie(CONTENT_TYPE_COOKIE_KEY);
		}

		let gspResult;
		if (gspFn) {
			gspResult = await gspFn(ctx);
		} else {
			gspResult = {
				props: {},
			};
		}
		if ('props' in gspResult) {
			return {
				...gspResult,
				props: {
					...gspResult.props,
					req,
					res,
				},
			};
		}

		// redirect/not found will be treated as default behavior
		return gspResult;
	};
}

export namespace action {
	export const getApiConfig = (customConfig = {} as PageConfig) => ({
		api: {
			...(customConfig.api ?? {}),
			bodyParser: false,
		},
	});

	export interface ActionWrapperOptions {
		fn: NextApiHandler,
		deserializers?: EncTypeDeserializerMap,
		/**
		 * Maps the Location header from the handler response to an accessible URL.
		 * @param url
		 */
		mapLocationToRedirectDestination?: (referer: string, url: string) => string,
	}

	export const wrapApiHandler = (
		options: ActionWrapperOptions,
	): NextApiHandler => async (req, res) => {
		const body = await deserializeBody({
			req,
			deserializers: options.deserializers,
		});
		const reqMut = req as unknown as Record<string, unknown>;
		reqMut.body = body;
		return options.fn(reqMut as unknown as DefaultNextApiRequest, res);
	};

	export const getServerSideProps = (
		options: ActionWrapperOptions,
	): GetServerSideProps => async (ctx) => {
		const { referer = '/' } = ctx.req.headers;

		const mockReq = {
			...ctx.req,
			body: await deserializeBody({
				req: ctx.req,
				deserializers: options.deserializers,
			}),
		} as DefaultNextApiRequest;

		const mockRes = new IceformNextServerResponse(ctx.req);
		await options.fn(mockReq, mockRes);

		const cookieManager = new CookieManager(ctx);
		cookieManager.setCookie(STATUS_CODE_COOKIE_KEY, mockRes.statusCode.toString());
		cookieManager.setCookie(STATUS_MESSAGE_COOKIE_KEY, mockRes.statusMessage);
		if (mockRes.data) {
			cookieManager.setCookie(BODY_COOKIE_KEY, mockRes.data as string);
			if (mockRes.contentType) {
				cookieManager.setCookie(CONTENT_TYPE_COOKIE_KEY, mockRes.contentType);
			}
		}
		const preventRedirect = (
			typeof mockReq.body === 'object'
			&& mockReq.body !== null
			&& PREVENT_REDIRECT_FORM_KEY in mockReq.body
		);
		const redirectDestination = (
			mockRes.location
			&& typeof options.mapLocationToRedirectDestination === 'function'
			&& !preventRedirect
		)
			? options.mapLocationToRedirectDestination(referer, mockRes.location)
			: referer;

		return {
			redirect: {
				destination: redirectDestination,
				statusCode: ACTION_STATUS_CODE,
			},
			props: {
				query: ctx.query,
				body: mockRes.data,
			},
		};
	};
}
